@@ -2,7 +2,7 @@ package Dancer::Plugin::Auth::Google;
 use strict;
 use warnings;
 
-our $VERSION = 0.06;
+our $VERSION = 0.07;
 
 use Dancer ':syntax';
 use Dancer::Plugin;
@@ -21,6 +21,7 @@ my $access_type;
 my $callback_url;
 my $callback_success;
 my $callback_fail;
+my $legacy_gplus;
 my $furl;
 
 register 'auth_google_init' => sub {
@@ -33,6 +34,7 @@ register 'auth_google_init' => sub {
     $callback_success = $config->{callback_success} || '/';
     $callback_fail    = $config->{callback_fail}    || '/fail';
     $access_type      = $config->{access_type}      || 'online';
+    $legacy_gplus     = $config->{legacy_gplus}     || 0;
 
     foreach my $param ( qw(client_id client_secret callback_url) ) {
         Carp::croak "'$param' is expected but not found in configuration"
@@ -55,13 +57,13 @@ register 'auth_google_authenticate_url' => sub {
     Carp::croak 'auth_google_init() must be called first'
         unless defined $callback_url;
 
-    my $uri = URI->new('https://accounts.google.com/o/oauth2/auth');
+    my $uri = URI->new('https://accounts.google.com/o/oauth2/v2/auth');
     $uri->query_form(
-        response_type => 'code',
         client_id     => $client_id,
         redirect_uri  => $callback_url,
         scope         => $scope,
         access_type   => $access_type,
+        response_type => 'code',
     );
 
     debug "google auth uri: $uri";
@@ -77,7 +79,7 @@ get '/auth/google/callback' => sub {
     return redirect $callback_fail unless $code;
 
     my $res = $furl->post(
-        'https://accounts.google.com/o/oauth2/token',
+        'https://www.googleapis.com/oauth2/v4/token',
         [ 'Content-Type' => 'application/x-www-form-urlencoded' ],
         {
             code          => $code,
@@ -89,39 +91,68 @@ get '/auth/google/callback' => sub {
     );
 
     my ($data, $error) = _parse_response( $res->decoded_content );
-    return send_error($error) if $error;
-
-    return send_error 'google auth: no access token present'
-        unless $data->{access_token};
+    if (ref $data && !$error) {
+        # Google tells us to ignore any unrecognized fields
+        # included in the response (like their "id_token").
+        $data = {
+            access_token  => $data->{access_token},
+            expires_in    => $data->{expires_in},
+            token_type    => $data->{token_type},
+            refresh_token => $data->{refresh_token},
+        };
+    }
+    else {
+        return send_error('google auth: ' . (defined $error ? $error : 'unknown error'));
+    }
 
     $res = $furl->get(
-        'https://www.googleapis.com/plus/v1/people/me',
+        'https://www.googleapis.com/oauth2/v2/userinfo',
         [ 'Authorization' => 'Bearer ' . $data->{access_token} ],
     );
 
     my $user;
     ($user, $error)  = _parse_response( $res->decoded_content );
-    return send_error($error) if $error;
+    return send_error("google auth: $error") if $error;
 
-    # we need to stringify our JSON::Bool data as some
-    # session backends might have trouble storing objects.
-    # we should be able to safely remove this once
-    # https://github.com/PerlDancer/Dancer-Session-Cookie/pull/1
-    # (or a similar solution) is merged.
-    if (exists $user->{image} and exists $user->{image}{isDefault}) {
-        $user->{image}{isDefault} = "$user->{image}{isDefault}";
-    }
-    if (exists $user->{isPlusUser}) {
-        $user->{isPlusUser} = "$user->{isPlusUser}";
-    }
-    if (exists $user->{verified}) {
-        $user->{verified} = "$user->{verified}";
+    if (exists $user->{verified_email}) {
+        # we stringify our JSON::Bool data as some session
+        # backends might have trouble storing objects.
+        $user->{verified_email} = "$user->{verified_email}";
     }
+    $user = _convert_to_legacy_gplus_format($user) if $legacy_gplus;
 
     session 'google_user' => { %$data, %$user };
     redirect $callback_success;
 };
 
+sub _convert_to_legacy_gplus_format {
+    my ($user) = @_;
+
+    return {
+        kind        => "plus#person",
+        displayName => $user->{name},
+        name => {
+            givenName => $user->{given_name},
+            familyName => $user->{family_name},
+        },
+        language   => $user->{locale},
+        isPlusUser => ($user->{link} && index($user->{link},'http') == 0 ? 1 : 0),
+        url        => $user->{link},
+        gender     => $user->{gender},
+        image => {
+            url => $user->{picture},
+            isDefault => 0,
+        },
+        domain         => $user->{hd},
+        emails         => [ { type => "account", value => $user->{email} } ],
+        etag           => undef,
+        verified       => $user->{verified_email},
+        circledByCount => undef,
+        id             => $user->{id},
+        objectType     => "person",
+    };
+}
+
 sub _parse_response {
     my ($response) = @_;
     my ($data, $error);
@@ -154,17 +185,17 @@ Dancer::Plugin::Auth::Google - Authenticate with Google
 
     auth_google_init;  # <-- don't forget to call this first!
 
-    before sub {
-        return unless request->path_info !~ m{^/auth/google/callback};
-        redirect auth_google_authenticate_url unless session('google_user');
-    };
-
     get '/' => sub {
-        "welcome, " . session('google_user')->{displayName}
+        if (session('google_user')) {
+            return 'you are logged in, ' . session('google_user')->{name};
+        }
+        else {
+            return redirect auth_google_authenticate_url;
+        }
     };
 
     get '/fail' => sub {
-        "oh noes!"
+        "Oh, noes! Your authentication failed :("
     };
 
 
@@ -193,17 +224,28 @@ for new ones.
 
 Anyone with a valid Google account can register an application. Go to
 L<http://console.developers.google.com>, then select a project or create
-a new one. After that, in the sidebar on the left, select "Credentials"
-under the "APIs and auth" option. In the "OAuth" section of that page,
-select B<Create New Client ID>. A dialog will appear.
+a new one. After that, in the sidebar on the left, select "Credentials".
+
+First, go to the I<OAuth consent screen> tab and set it up with you website's
+logo, desired credentials (the "email" and "profile" ones are granted
+by default) and, specially, your B<authorized domains>. We'll need those for
+the next step!
+
+Now go to the I<Credentials> tab and click the B<Create credentials>
+button/dropdown and select B<OAuth client ID>.
 
 =for HTML
 <p><img src="https://raw.githubusercontent.com/garu/Dancer-Plugin-Auth-Google/master/share/create-new-id.png"></p>
 
-In the "Application type" section of the dialog, make sure you select
-"Web application". In the "Authorized JavaScript origins" field, make
-sure you put the domains of both your development server and your
-production one (e.g.: http://localhost:3000 and http://mywebsite.com).
+A dialog will appear. In the "Application type" section of the dialog,
+select I<"Web application">.
+
+Under the "Authorized JavaScript origins" field, put the domains of both
+your development server and your production one
+(e.g.: http://localhost:3000 and http://mywebsite.com). You will only be
+able to include domains listed under your I<authorized domain list>,
+which you set on the previous step (though localhost domains are ok).
+
 Same thing goes for the "Redirect URIs": those B<**MUST**> be the same
 as you will set in your app and Google won't redirect to any page that
 is not listed (don't worry, you can edit this later too).
@@ -213,15 +255,12 @@ is not listed (don't worry, you can edit this later too).
 
 Again, make sure the "Redirect URIs" contains both your development
 url (e.g. C<http://localhost:3000/auth/google/callback>) and production
-(e.g. C<http://mycoolwebsite.com/auth/google/callback>).
+(e.g. C<http://mywebsite.com/auth/google/callback>). It's usually a good
+practice to add I<both> HTTP and HTTPS callback urls.
 
 After you're finished, copy the "Client ID" and "Client Secret" data
 of your newly created app. It should be listed on that same panel
-(you can check it anytime by going to the "Credentials" option under
-"APIs & auth":
-
-=for HTML
-<p><img src="https://raw.githubusercontent.com/garu/Dancer-Plugin-Auth-Google/master/share/client-id.png"></p>
+(you can check it anytime by going to the "Credentials" option)
 
 =head2 Configuration
 
@@ -240,6 +279,7 @@ Plugins / Auth::Google, like so:
             callback_url:     'http://localhost:3000/auth/google/callback'
             callback_success: '/'
             callback_fail:    '/fail'
+            legacy_gplus:     0
 
 Of those, only "client_id", "client_secret" and "callback_url" are mandatory.
 If you omit the other ones, they will assume their default values, as listed
@@ -247,7 +287,7 @@ above.
 
 Specifically, it is a good practice to change the C<callback_url> depending on
 whether you're on a development or production environment. Dancer makes this
-easier for you by letting you split your settings, leaving the basic plugin
+trivial by letting you split your settings, leaving the basic plugin
 settings on C<config.yml> and specific C<callback_url> definitions on
 C<environments/development.yml> and C<environments/production.yml>:
 
@@ -269,13 +309,13 @@ And
 Since this plugin is meant mainly for authentication, the default scope
 is 'profile'. That should give you general profile data for the user, such
 as full name, id, profile url, etc. See
-L<https://developers.google.com/+/api/oauth#login-scopes> for available
+L<https://developers.google.com/identity/protocols/googlescopes> for available
 scopes to chose from. You can set as many as you like, separated by space.
 A usual combination is 'profile email'. If you want a Google-specific scope
 (i.e. those with a "." in the name) make sure you add the full URL as
-specified in the document above. For example, the proper way to ask for a
-user's social features is not "plus.login", but
-"https://www.googleapis.com/auth/plus.login".
+specified in the link above. For example, the proper way to ask for
+restricted Google Drive access is "https://www.googleapis.com/auth/drive.file",
+not just "drive.file".
 
 
 =head1 EXPORTS
@@ -290,10 +330,34 @@ read your configuration and create everything that it needs.
 
 =head2 auth_google_authenticate_url
 
-This function returns an authorize URI for redirecting unauthenticated
-users. You should use this in a before filter like the "synopsis"
+This function returns an authorization URI for redirecting unauthenticated
+users. You should use this in a before filter like the "SYNOPSIS"
 demo above.
 
+Google allows you to send additional C<key=value> data to preserve state
+in your application throughout the OAuth2 process, via the I<state> variable.
+To do so, simply tweak the URI object with that additional info:
+
+    use URI::Escape;
+
+    if (!session('google_user')) {
+        my $state = 'CSRF=my-special-token&other=whatever';
+
+        my $uri = auth_google_authenticate_url;
+        $uri->query_form(
+            $uri->query_form, # <-- required so we ADD instead of REPLACE
+            state => URI::Encode->new->encode($state),
+        );
+
+        return redirect $uri;
+    }
+
+If you do this, the "state" data will be sent back to you on the callback
+as a request parameter. You can then use it to direct your user to the
+proper resource or check it against forgery attacks - assuming you also
+stored it in a session variable of some sort.
+
+
 =head1 ROUTE HANDLERS
 
 The plugin defines the following route handler automatically:
@@ -301,7 +365,7 @@ The plugin defines the following route handler automatically:
 =head2 /auth/google/callback
 
 This route handler is responsible for catching back a user that has just
-authenticated herself with Google's OAuth. The route handler saves tokens
+authenticated herself with Google's OAuth2. The route handler saves tokens
 and user information in the session and then redirects the user to the URI
 specified by callback_success.
 
@@ -309,20 +373,62 @@ If the validation of the token returned by Google failed or was denied,
 the user will be redirected to the URI specified by callback_fail. Otherwise,
 this route will point the user to callback_success.
 
+=head3 What data is under session('google_user')?
+
+After getting basic authentication data, this module uses the Google Identity
+Platform API, to fetch basic user profile data, like a unique id, email, name
+and picture. C<< session('google_user') >> looks like so (random hash order!):
+
+    {
+        access_token   => 'THIS IS THE ACCESS TOKEN TO USER',
+        expires_in     => 3920,  # seconds!
+        token_type     => 'Bearer',
+        refresh_token  => 'USE THIS TO REFRESH ACCESS WITHOUT NEW AUTH',
+        name           => "FirstName LastName",
+        given_name     => "FirstName",
+        family_name    => "LastName",
+        picture        => "https://SOME-GOOGLE-URL/photo.jpg",
+        locale         => "en",
+        gender         => "female", # or "male", or "other"
+        email          => "email@example.com",
+        link           => "https://OLD-GOOGLE-PLUS-PROFILE-THIS-MIGHT-GO-AWAY",
+        id             => "NUMERIC UNIQUE USER ID HERE",
+        hd             => "HOSTED DOMAIN",
+        verified_email => 1,
+    }
+
+=head4 NOTE: G+ is no more. Add 'legacy_gplus: 1' to keep old code running.
+
+Up to version 0.06 of this module the C<< session('google_user') >>
+data structure was as returned by Google Plus' API. Google decided to
+discontinue G+ and its API on March 7th 2019, so since version 0.07
+we fetch user information from Google's C<oauth2/v2/userinfo> endpoint.
+
+Those two structures are very different, so
+B<you will need to update your code> if you used any version of this module
+prior to 0.07. If you don't want to, add the C<legacy_gplus> option to
+your configuration with a true value and C<< session('google_user') >>
+will instead return a data structure as closely matched as possible to
+the old version. Note that some data may not be available anymore
+(particularly, I<circleByCount> and I<etag>), in which case the structure
+will return C<undef> - but you tested those fields already, right?
+
+
 =head1 ACCESSING OTHER GOOGLE APIS
 
-Once the user is authenticated, your session data will contain the access
-token:
+As noted above, once the user is authenticated your session data
+will contain the access token:
 
     my $token = session('google_user')->{access_token};
 
 You can use that access token to make calls to a Google API on behalf of
-the user. See L<https://developers.google.com/accounts/docs/OAuth2WebServer>
+the user (provided the user authorized the related scope.
+See L<https://developers.google.com/identity/protocols/OAuth2WebServer>
 for more information on this.
 
 =head1 BUGS
 
-Please submit any bug reports or feature requests either on RT or Github.
+Please submit any bug reports or feature requests to Github.
 
 =head1 ACKNOWLEDGEMENTS
 
@@ -331,7 +437,7 @@ Dancer::Plugin::Auth::Twitter and Dancer::Plugin::Auth::Facebook.
 
 =head1 COPYRIGHT AND LICENCE
 
-Copyright (C) 2014, Breno G. de Oliveira
+Copyright (C) 2014-2019, Breno G. de Oliveira
 
 This library is free software; you can redistribute it and/or modify
 it under the same terms as Perl itself.
